import re
import inspect
import logging
import warnings
from .y4a_telegram import send_message

warnings.filterwarnings("ignore", category=UserWarning)

logging.basicConfig(
    level=logging.INFO,
    format='%(asctime)s - %(levelname)s - %(message)s',
    datefmt='%Y-%m-%d %H:%M:%S',
)


def telegram_alert(context) -> None:
    """
    Airflow DAG failure notification to Telegram
    """

    owner = '@' + context['dag'].default_args['owner']

    message = f"\u26d4 *{context['task_instance']}*\n"
    message += f"Owner: {owner}\n"
    message += f"DAG ID: *{context['task_instance'].dag_id}*\n"
    message += f"Task ID: *{context['task_instance'].task_id}*\n"
    message += f"Execution Time: *{context['task_instance'].execution_date}*\n"
    message += f"Error: *{context['exception']}*\n"
    log_url = str(context['task_instance'].log_url)
    log_url = log_url.replace('localhost', 'af-in-master-sop.yes4all.internal')
    message += f"Log URL: [Link]({log_url})"

    telegram_token = "6318613524:AAG3_JGEsTZbSvcupG5aJk-jZPzghuf3yZ4"
    chat_id = "-868321875"

    send_message(
        text=message,
        bot_token=telegram_token,
        chat_id=chat_id,
    )


def auto_dag_id() -> str:
    """
    Auto naming for the Airflow DAG ID based on the file directory

    :return: name of the Airflow DAG ID
    """

    current_frame = inspect.currentframe()
    current_path = inspect.getfile(current_frame.f_back)
    airflow_path = current_path.split('/airflow/dags/')[-1]
    airflow_path = airflow_path.replace('sop_ingestion', 'de')
    dag_id = airflow_path\
        .replace('.py', '')\
        .lower().strip()
    dag_id = re.sub(
        r'[^a-zA-Z0-9/]+',
        '_',
        dag_id,
    )
    dag_id = dag_id.replace('/', '.')

    return dag_id



import datetime
import logging
import pytz
import warnings
import pandas as pd
from psycopg2 import connect
from ..y4a_credentials import get_credentials

logging.basicConfig(
    level=logging.INFO,
    format='%(asctime)s - %(levelname)s - %(message)s',
    datefmt='%Y-%m-%d %H:%M:%S',
)

warnings.filterwarnings("ignore", category=UserWarning)


class MonitorUserExternal:
    def __init__(
        self,
        conn_username: str,
        conn_password: str,
        conn_host: str,
        conn_db: str,
        conn_number: int,
    ) -> None:
        self.conn_username = conn_username
        self.conn_password = conn_password
        self.conn_host = conn_host
        self.conn_db = conn_db
        self.conn_number = conn_number

        self.log_conn_df = pd.DataFrame(
            {
                'time_conn': str(
                    datetime.datetime.now(
                        pytz.timezone('Asia/Ho_Chi_Minh'),
                    ).replace(microsecond=0).replace(tzinfo=None)
                ),
                'host': self.conn_host,
                'username': self.conn_username,
                'password': self.conn_password,
                'database': self.conn_db,
                'num_conn': self.conn_number,
            },
            index=[0],
        )

        self.id_conn = None

        self.credentials = get_credentials(
            platform='pg',
            account_name='da_admin',
        )

    def execute(
        self,
        sql: str,
        fetch_output: bool = False,
    ) -> list | None:
        output = None

        conn = connect(
            user=self.credentials['user_name'],
            password=self.credentials['password'],
            host='172.30.12.153',
            port=5432,
            database='serving',
        )

        try:
            cur = conn.cursor()

            cur.execute(sql)

            if fetch_output:
                output = cur.fetchall()

            cur.close()
            conn.commit()
            conn.close()
        except Exception as e:
            logging.error(e)
            conn.close()

        return output

    def generate_log_conn(self) -> None:
        sql = "INSERT INTO metadata.monitor_user_sql_connection "
        sql += "(time_conn, host, username, password, database, num_conn) "
        sql += f"VALUES ('{self.log_conn_df['time_conn'].values[0]}', "
        sql += f"'{self.log_conn_df['host'].values[0]}', "
        sql += f"'{self.log_conn_df['username'].values[0]}', "
        sql += f"'{self.log_conn_df['password'].values[0]}', "
        sql += f"'{self.log_conn_df['database'].values[0]}', "
        sql += f"{self.log_conn_df['num_conn'].values[0]}) "
        sql += "RETURNING id_conn"

        self.id_conn = self.execute(
            sql=sql,
            fetch_output=True,
        )[0][0]

    def update_log_conn(self) -> None:
        sql = "UPDATE metadata.monitor_user_sql_connection SET "
        sql += "is_closed = 1 WHERE "
        sql += f"id_conn = {self.id_conn} "
        sql += f"AND time_conn = '{self.log_conn_df['time_conn'].values[0]}' "
        sql += f"AND host = '{self.log_conn_df['host'].values[0]}' "
        sql += f"AND username = '{self.log_conn_df['username'].values[0]}' "

        self.execute(sql)

    def grant_table(
        self,
        schema: str,
        table: str,
        list_users: list,
        privileges: list = ['SELECT'],
        all_privileges: bool = False,
    ):
        conn = connect(
            user=self.credentials['user_name'],
            password=self.credentials['password'],
            host=self.conn_host,
            port=5432,
            database=self.conn_db,
        )

        try:
            cur = conn.cursor()

            if 'da' in list_users:
                list_users = [
                    get_credentials(
                        platform='pg',
                        account_name=user,
                    )['user_name'] for user in [
                        'linhvk', 'giangpt', 'hieulm',
                        'huongtmt', 'pi', 'toannh',
                        'trieuna', 'hieulm', 'hoaidtm',
                        'duikha', 'longgh', 'huypv',
                        'huytln', 'quandm', 'nguyentnt',
                        'clong',
                    ]
                ]
            else:
                list_users = [
                    get_credentials(
                        platform='pg',
                        account_name=user,
                    )['user_name'] for user in list_users
                ]

            for user in list_users:
                schema_grant = f"GRANT USAGE ON SCHEMA {schema} TO {user}"
                cur.execute(schema_grant)
                logging.info(f'Grant schema usage for {user}')

                if all_privileges:
                    table_grant = f"GRANT ALL PRIVILEGES ON {schema}.{table} "
                    table_grant += f"to {user}"
                else:
                    table_privileges = ', '.join(privileges)
                    table_grant = f"GRANT {table_privileges} ON "
                    table_grant += f"{schema}.{table} to {user}"

                cur.execute(table_grant)
                logging.info(table_grant)

            cur.close()
            conn.commit()
            conn.close()
        except Exception as e:
            logging.error(e)
            conn.close()



import logging
import warnings
from typing import Callable
import pandas as pd
import numpy as np
import gspread
from ..y4a_credentials import get_credentials

warnings.filterwarnings("ignore", category=UserWarning)

logging.basicConfig(
    level=logging.INFO,
    format='%(asctime)s - %(levelname)s - %(message)s',
    datefmt='%Y-%m-%d %H:%M:%S',
)


class GGSheetUtils:
    """
    Utils for Google Sheets

    :param account_name: the client account name for Google Sheet
        defaults to 'da'
    :param auth_dict: the dictionary of user credentials for the Google Sheet,
        defaults to None
        if None, the module will use the default credentials
        of DA team
    """

    def __init__(
        self,
        account_name: str = 'da',
        user_creds: dict = None,
    ) -> None:
        self.account_name = account_name

        if user_creds:
            self.auth_dict = user_creds
        else:
            self.credentials = get_credentials(
                platform='gg_api',
                account_name=self.account_name,
            )

            self.email = self.credentials['email']
            self.private_key = self.credentials['secret_key']

            self.auth_dict = {
                'client_email': self.email,
                'private_key': self.private_key,
                'token_uri': 'https://oauth2.googleapis.com/token',
            }

            self.client = gspread.service_account_from_dict(self.auth_dict)

    def open_spread_sheet(
        self,
        sheet_id: str,
    ) -> Callable:
        spread_sheet = self.client.open_by_key(
            key=sheet_id,
        )

        return spread_sheet

    def get_work_sheet(
        self,
        spread_sheet: Callable,
        sheet_name: str,
    ) -> Callable:
        work_sheet = spread_sheet.worksheet(sheet_name)

        return work_sheet

    def create_spread_sheet(
        self,
        sheet_name: str,
        folder_id: str = None,
        share_to: list = [],
    ) -> str:
        """
        Create a new spread sheet

        :param sheet_name: name of the sheet
        :param folder_id: id of the folder contains spreadsheet
        :param share_to: list of email to share the spreadsheet
            defaults to []

        :return: the created spreadsheet id
        """

        spread_sheet = self.client.create(
            title=sheet_name,
            folder_id=folder_id,
        )
        if share_to:
            for mail in share_to:
                spread_sheet.share(
                    email_address=mail,
                    perm_type='user',
                    role='writer',
                )

        return spread_sheet.id

    def list_all_work_sheets(
        self,
        sheet_id: str,
    ) -> list:
        """
        Get all available worksheet of spreadsheet

        :param sheet_id: spreadsheet id

        :return: list all worksheets of spreadsheet
        """

        spread_sheet = self.open_spread_sheet(sheet_id)

        work_sheets = spread_sheet.worksheets()

        return work_sheets

    def delete_work_sheet(
        self,
        sheet_id: str,
        sheet_name: str = 'Sheet1',
    ) -> None:
        """
        Delete specific worksheet of spreadsheet

        :param sheet_id: spreadsheet id
        :param sheet_name: worksheet name
            defaults to 'Sheet1'

        """

        spread_sheet = self.open_spread_sheet(sheet_id)

        work_sheet = self.get_work_sheet(
            spread_sheet=spread_sheet,
            sheet_name=sheet_name,
        )

        spread_sheet.del_worksheet(work_sheet)

    def clear_work_sheet(
        self,
        sheet_id: str,
        sheet_name: str = 'Sheet1',
    ) -> None:
        """
        Clear all data of specific worksheet of spreadsheet

        :param sheet_id: spreadsheet id
        :param sheet_name: worksheet name
            defaults to 'Sheet1'

        """

        spread_sheet = self.open_spread_sheet(sheet_id)

        work_sheet = self.get_work_sheet(
            spread_sheet=spread_sheet,
            sheet_name=sheet_name,
        )

        work_sheet.clear()

    def get_data(
        self,
        sheet_id: str,
        sheet_name: str = 'Sheet1',
        range_from: str = 'A',
        range_to: str = 'Z',
        columns_first_row: bool = False,
    ) -> pd.DataFrame:
        """
        Get data from the given sheet

        :param sheet_id: spreadsheet name
        :param sheet_name: worksheet name
            defaults to 'Sheet1'
        :param range_from: the begining of the range
            of data from sheet to get
            defaults to 'A'
        :param range_to: the end of the range
            of data from sheet to get
            defaults to 'Z'
        :param columns_first_row: whether to convert the first row
            to columns
            defaults to False

        :return: the dataframe contains data from sheet
        """

        spread_sheet = self.open_spread_sheet(sheet_id)

        data_range = f'{range_from}:{range_to}'
        if sheet_name:
            data_range = f'{sheet_name}!{data_range}'

        data = spread_sheet.values_get(data_range)['values']
        df = pd.DataFrame(data)
        if columns_first_row:
            df.columns = df.iloc[0].to_list()
            df = df.iloc[1:].reset_index(drop=True)

        return df

    def insert_data(
        self,
        data: pd.DataFrame,
        sheet_id: str,
        sheet_name: str = 'Sheet1',
        from_row_index: int = 1,
        insert_column_names: bool = False,
    ) -> None:
        """
        Insert data to the given sheet

        :param data: dataframe contains data to insert
        :param sheet_id: spreadsheet id
        :param sheet_name: worksheet name
            defaults to 'Sheet1'
        :param from_row_index: the index of the row
            beginning to insert
            defaults to 1
        :param insert_column_names: whether to insert column names
            defaults to False
        """

        spread_sheet = self.open_spread_sheet(sheet_id)

        work_sheet = self.get_work_sheet(
            spread_sheet=spread_sheet,
            sheet_name=sheet_name,
        )

        data_values = data.replace(
            np.nan,
            None,
        ).values.tolist()

        if insert_column_names:
            col_values = [data.columns.to_list()]
            work_sheet.insert_rows(
                col_values,
                row=from_row_index,
            )
            work_sheet.insert_rows(
                data_values,
                row=from_row_index+1,
            )
        else:
            work_sheet.insert_rows(
                data_values,
                row=from_row_index,
            )

    def update_data(
        self,
        data: pd.DataFrame,
        sheet_id: str,
        sheet_name: str = 'Sheet1',
        range_from: str = 'A',
        range_to: str = 'Z',
    ) -> None:
        """
        Update data of the given sheet

        :param data: dataframe contains data to update
        :param sheet_id: spreadsheet name
        :param sheet_name: worksheet name
            defaults to 'Sheet1'
        :param range_from: the begining of the range
            of data from sheet to update
            defaults to 'A'
        :param range_to: the end of the range
            of data from sheet to update
            defaults to 'Z'
        """

        spread_sheet = self.open_spread_sheet(sheet_id)

        work_sheet = self.get_work_sheet(
            spread_sheet=spread_sheet,
            sheet_name=sheet_name,
        )

        data_values = data.replace(
            np.nan,
            None,
        ).values.tolist()

        work_sheet.update(
            f'{range_from}:{range_to}',
            data_values,
        )

    def remove_data(
        self,
        sheet_id: str,
        sheet_name: str = 'Sheet1',
        list_range: list = [
            'A1:Z1',
            'A4:Z4',
        ],
    ) -> None:
        """
        Remove data from specific range of the given sheet

        :param sheet_id: spreadsheet name
        :param sheet_name: worksheet name
            defaults to 'Sheet1'
        :param list_range: list of data ranges to remove
            defaults to ['A1:Z1', 'A4:Z4']
        """

        spread_sheet = self.open_spread_sheet(sheet_id)

        work_sheet = self.get_work_sheet(
            spread_sheet=spread_sheet,
            sheet_name=sheet_name,
        )
        work_sheet.batch_clear(list_range)



import logging
import warnings
from ..gg_api.y4a_sheet import GGSheetUtils
from ..datalake.y4a_minio import MinioUtils
from ..sql.y4a_postgresql import PostgreSQLUtils
from ..sql.y4a_mysql import MySQLUtils

logging.basicConfig(
    level=logging.INFO,
    format='%(asctime)s - %(levelname)s - %(message)s',
    datefmt='%Y-%m-%d %H:%M:%S',
)

warnings.filterwarnings("ignore", category=UserWarning)


class DAConfig:
    def __init__(
        self,
        account_name: str,
        **kwargs,
    ) -> None:
        # Google Sheet utils
        self.sheet_utils = GGSheetUtils(
            account_name='da',
        )

        # MinIO utils
        self.minio_utils = MinioUtils(
            account_name='sop',
        )

        # PostgreSQL utils
        self.pg_raw_r_utils = PostgreSQLUtils(
            account_name=account_name,
            db_host='raw_repl',
            db='raw',
        )
        self.pg_raw_w_utils = PostgreSQLUtils(
            account_name=account_name,
            db_host='raw_master',
            db='raw',
        )
        self.pg_serving_r_utils = PostgreSQLUtils(
            account_name=account_name,
            db_host='serving_repl',
            db='serving',
        )
        self.pg_serving_w_utils = PostgreSQLUtils(
            account_name=account_name,
            db_host='serving_master',
            db='serving',
        )

        # MySQL utils
        self.mysql_utils = MySQLUtils(
            account_name='da',
        )

        for key, value in kwargs.items():
            setattr(self, key, value)



import logging
import warnings
from io import BytesIO
from minio import Minio
import pandas as pd
from ..y4a_credentials import get_credentials

warnings.filterwarnings("ignore", category=UserWarning)

logging.basicConfig(
    level=logging.INFO,
    format='%(asctime)s - %(levelname)s - %(message)s',
    datefmt='%Y-%m-%d %H:%M:%S',
)


class MinioUtils:
    """
    Utils for MinIO storage

    :param account_name: the client account name to minio storage
    """
    def __init__(
        self,
        account_name: str = 'sop',
    ) -> None:
        self.account_name = account_name
        self.credentials = get_credentials(
            platform='minio',
            account_name=self.account_name,
        )
        self.access_key = self.credentials['access_key']
        self.secret_key = self.credentials['secret_key']

        self.client = Minio(
            endpoint='storage-api.yes4all.internal',
            access_key=self.access_key,
            secret_key=self.secret_key,
            secure=False,
        )

    def data_exist(
        self,
        mode: str,
        file_path: str,
        bucket_name: str = 'sc-bucket',
    ) -> bool:
        """
        Check if the data file exists in a directory

        :param mode: the data storage mode
            the value is either 'prod' or 'stag'
        :param file_path: the directory to check
            notes that the path is considered after
            the data storage mode parent directory
            and remove the ".parquet" extension
        :param bucket_name: the name of the bucket to check
            defaults to 'sc-bucket'

        :return: True if the data file exists
            otherwise False
        """

        obj = self.client.list_objects(
            bucket_name=bucket_name,
            prefix=f'{mode}{file_path}.parquet'
        )
        obj = [*obj]

        if len(obj) == 0:
            return False
        return True

    def get_data_value_exist(
        self,
        mode: str,
        file_path: str,
        column_key: str,
        bucket_name: str = 'sc-bucket',
    ) -> list:
        """
        Get the distinct values of a specified column
            of data in the data directory

        :param mode: the data storage mode
            the value is either 'prod' or 'stag'
        :param file_path: the directory to get distinct values
            notes that the path is considered after
            the data storage mode parent directory
            and remove the ".parquet" extension
        :param column_key: the column name to get distinct values
        :param bucket_name: the name of the bucket
            to get distinct values
            defaults to 'sc-bucket'

        :return: list of distinct values
        """

        df = self.get_data_wildcard(
            mode,
            file_path,
            bucket_name,
        )

        if len(df) == 0:
            return list()

        return list(df[column_key].unique())

    def load_data(
        self,
        data: pd.DataFrame,
        mode: str,
        file_path: str,
        bucket_name: str = 'sc-bucket',
    ) -> None:
        """
        Load data from dataframe to storage

        :param data: dataframe contains data to load
        :param mode: the data storage mode
            the value is either 'prod' or 'stag'
        :param file_path: the directory to load the data
            notes that the path is considered after
            the data storage mode parent directory
            and remove the ".parquet" extension
        :param bucket_name: the name of the bucket
            to load the data
            defaults to 'sc-bucket'
        """

        if mode not in ['prod', 'stag']:
            logging.error('Mode must be prod or stag')
            return

        parquet_data = data.to_parquet(index=False)
        bytes_data = BytesIO(parquet_data)

        self.client.put_object(
            bucket_name=bucket_name,
            object_name=f'{mode}{file_path}.parquet',
            data=bytes_data,
            length=bytes_data.getbuffer().nbytes,
            content_type=f'{mode}/parquet',
        )

        logging.info('Done loading data to Minio storage')

    def get_data(
        self,
        mode: str,
        file_path: str,
        bucket_name: str = 'sc-bucket',
    ) -> pd.DataFrame:
        """
        Get data from a single file of directory of storage

        :param mode: the data storage mode
            the value is either 'prod' or 'stag'
        :param file_path: the directory to get the data
            notes that the path is considered after
            the data storage mode parent directory
            and remove the ".parquet" extension
        :param bucket_name: the name of the bucket
            to get the data
            defaults to 'sc-bucket'

        :return: dataframe contains data to get
        """

        if mode not in ['prod', 'stag']:
            logging.error('Mode must be prod or stag')
            return

        parquet_data = BytesIO(
            self.client.get_object(
                bucket_name=bucket_name,
                object_name=f'{mode}{file_path}.parquet',
            ).data
        )
        df = pd.read_parquet(parquet_data)

        return df

    def get_data_wildcard(
        self,
        mode: str,
        file_path: str,
        bucket_name: str = 'sc-bucket',
    ) -> pd.DataFrame:
        """
        Get data from multiple files of directories of storage

        :param mode: the data storage mode
            the value is either 'prod' or 'stag'
        :param file_path: the parent directory to get the data
            notes that the path is considered after
            the data storage mode parent directory
        :param bucket_name: the name of the bucket
            to get the data
            defaults to 'sc-bucket'

        :return: dataframe contains data to get
        """

        data = list()

        if mode not in ['prod', 'stag']:
            logging.error('Mode must be prod or stag')
            return

        objects = self.client.list_objects(
            bucket_name=bucket_name,
            prefix=f'{mode}{file_path}',
            recursive=True,
        )

        num_objects = 0
        for obj in objects:
            if '.parquet' in obj.object_name:
                file_name = obj.object_name.replace(
                    '.parquet', ''
                ).replace(
                    f'{mode}/', '/'
                )
                data.append(
                    self.get_data(
                        mode,
                        file_name,
                        bucket_name,
                    )
                )
                num_objects += 1
                if num_objects % 100 == 0:
                    logging.info(f'Got {num_objects} objects')

        if len(data) > 0:
            df = pd.concat(data).reset_index(drop=True)
            return df

        return pd.DataFrame()



import logging
import datetime
import warnings
import mysql.connector
import pandas as pd
from ..y4a_credentials import get_credentials

warnings.filterwarnings("ignore", category=UserWarning)

logging.basicConfig(
    level=logging.INFO,
    format='%(asctime)s - %(levelname)s - %(message)s',
    datefmt='%Y-%m-%d %H:%M:%S',
)


class MySQLUtils:
    def __init__(
        self,
        account_name: str,
    ) -> None:
        self.account_name = account_name
        self.credentials = get_credentials(
            platform='mysql',
            account_name=self.account_name,
        )
        self.db_user = self.credentials['user_name']
        self.db_password = self.credentials['password']
        self.db_host = 'sc-db.yes4all.internal'

    def _serialize_cell(
        self,
        cell,
    ):
        if cell is None:
            return None
        if isinstance(cell, datetime.datetime):
            return cell.isoformat()

        return str(cell)

    def connect(
        self,
        database: str,
    ):
        conn = mysql.connector.connect(
            host=self.db_host,
            user=self.db_user,
            password=self.db_password,
            database=database,
        )

        return conn

    def read_sql_file(
        self,
        sql_file_path: str,
    ):
        with open(sql_file_path, 'r') as file:
            sql = file.read()

        return sql

    def insert_data(
        self,
        data: pd.DataFrame,
        database: str,
        table: str,
        replace: bool = False,
        commit_every: int = 1000,
    ):
        if len(data) == 0:
            return

        conn = self.connect(database)
        cur = conn.cursor()

        target_fields = data.columns.to_list()
        target_fields = ", ".join(target_fields)
        target_fields = "({})".format(target_fields)

        rows = [
            tuple(r) for r in data.to_numpy()
        ]

        for i, row in enumerate(rows, 1):
            lst = []
            for cell in row:
                lst.append(self._serialize_cell(cell))
            values = tuple(lst)
            placeholders = ["%s", ] * len(values)
            if not replace:
                sql = "INSERT INTO "
            else:
                sql = "REPLACE INTO "
            sql += "{0} {1} VALUES ({2})".format(
                table,
                target_fields,
                ",".join(placeholders)
            )
            cur.execute(sql, values)
            if i % commit_every == 0:
                conn.commit()
                logging.info(
                    f"Loaded {i} into {table} rows so far"
                )

        conn.commit()
        conn.close()

        logging.info(
            f"Done loading. Loaded a total of {i} rows"
        )

    def update_table(
        self,
        data: pd.DataFrame,
        database: str,
        table: str,
        columns: list,
        primary_keys: list,
        commit_every: int = 1000,
    ):
        if len(data) == 0:
            return

        conn = self.connect(database)
        cur = conn.cursor()

        sql = f"UPDATE {table} SET "
        sql += ", ".join(
            [
                f"{col} = %s" for col in columns
            ]
        )
        sql += " WHERE "
        sql += " AND ".join(
            [
                f"{key} = %s" for key in primary_keys
            ]
        )

        sub_cols = columns + primary_keys
        rows = [
            tuple(r) for r in data[sub_cols].to_numpy()
        ]

        for i, row in enumerate(rows, 1):
            lst = []
            for cell in row:
                lst.append(self._serialize_cell(cell))
            values = tuple(lst)
            cur.execute(sql, values)
            if i % commit_every == 0:
                conn.commit()
                logging.info(
                    f"Updated {i} to {table} rows so far"
                )

        conn.commit()
        conn.close()

        logging.info(
            f"Done updating. Updated a total of {i} rows"
        )

    def get_data(
        self,
        sql: str,
        database: str,
    ):
        conn = self.connect(database)

        data = pd.read_sql(
            sql=sql,
            con=conn,
        )

        conn.close()

        return data

    def select_distinct(
        self,
        col: str,
        database: str,
        table: str,
    ):
        conn = self.connect(database)
        sql = f"SELECT DISTINCT {col} "
        sql += f"FROM {table}"

        data = pd.read_sql(
            sql=sql,
            con=conn,
        )[col].to_list()

        conn.close()

        return data

    def show_columns(
        self,
        database: str,
        table: str,
    ):
        columns = self.get_data(
            sql=f"SHOW COLUMNS FROM {table}",
            database=database,
        )['Field'].to_list()

        return columns

    def add_column(
        self,
        database: str,
        table: str,
        column_name: str,
        dtype: str,
    ):
        sql = f"ALTER TABLE {table} "
        sql += f"ADD COLUMN {column_name} {dtype}"

        self.execute(
            sql=sql,
            database=database,
        )

    def execute(
        self,
        sql: str,
        database: str,
    ):
        conn = self.connect(database)
        cur = conn.cursor()

        cur.execute(sql)

        conn.commit()

        cur.close()
        conn.close()



import logging
import datetime
import warnings
from typing import Callable
import pandas as pd
from psycopg2 import pool, connect
from io import StringIO
from .y4a_monitor_user import MonitorUserExternal
from ..y4a_retry import retry_on_error
from ..y4a_credentials import get_credentials

warnings.filterwarnings("ignore", category=UserWarning)

logging.basicConfig(
    level=logging.INFO,
    format='%(asctime)s - %(levelname)s - %(message)s',
    datefmt='%Y-%m-%d %H:%M:%S',
)


class PostgreSQLUtils:
    """
    Utils for PostgreSQL

    :param account_name: the client account name to connect to PostgreSQL
    :param db_host: host db name to connect
    :param db: PostgreSQL database to connect
        defaults to 'serving'
    """

    def __init__(
        self,
        account_name: str,
        db_host: str,
        db: str = 'serving',
    ) -> None:
        self.hosts = {
            'raw_master': '172.30.12.153',
            'raw_repl': '172.30.12.154',
            'serving_master': '172.30.12.153',
            'serving_repl': '172.30.12.154',
        }
        self.account_name = account_name
        self.credentials = get_credentials(
            platform='pg',
            account_name=self.account_name,
        )
        self.db_user = self.credentials['user_name']
        self.db_password = self.credentials['password']
        self.db_host = db_host
        self.host = self.hosts[self.db_host]
        self.db = db

        self.external_monitor = None

    @retry_on_error(max_retries=5)
    def create_pool_conn(
        self,
        pool_size: int = 1,
    ) -> Callable:
        """
        Create a connection pool to connect to PostgreSQL

        :param pool_size: number of connections in the pool

        :return: connection pool contains multiple connections
            to the database
        """

        db_pool_conn = pool.SimpleConnectionPool(
            minconn=1,
            maxconn=pool_size,
            user=self.db_user,
            password=self.db_password,
            host=self.host,
            port=5432,
            database=self.db,
        )
        self.generate_log_conn(pool_size)

        return db_pool_conn

    def close_pool_conn(
        self,
        db_pool_conn: Callable,
    ) -> None:
        """
        Close and remove the connection pool

        :param db_pool_conn: connection pool contains multiple connections
            to the database
        """

        self.update_log_conn()
        db_pool_conn.closeall()

    @retry_on_error(max_retries=5)
    def open_conn(self) -> Callable:
        """
        Create a new connection to PostgreSQL

        :return: connection object to connect to database
        """

        conn = connect(
            user=self.db_user,
            password=self.db_password,
            host=self.host,
            port=5432,
            database=self.db,
        )

        return conn

    def close_conn(
        self,
        conn: Callable,
    ) -> None:
        """
        Close the connection to PostgreSQL

        :param conn: connection object to connect to database
        """

        conn.close()

    def generate_log_conn(
        self,
        conn_number: int,
    ) -> None:
        try:
            self.external_monitor = MonitorUserExternal(
                conn_username=self.account_name,
                conn_password=self.db_password,
                conn_host=self.host,
                conn_db=self.db,
                conn_number=conn_number,
            )
            self.external_monitor.generate_log_conn()
        except Exception as e:
            logging.error('Can not track log for this connection')
            logging.error(e)

    def update_log_conn(self) -> None:
        try:
            self.external_monitor.update_log_conn()
        except Exception as e:
            logging.error('Can not update log for this connection')
            logging.error(e)

    def grant_table(
        self,
        schema: str,
        table: str,
        list_users: list,
        privileges: list = ['SELECT'],
        all_privileges: bool = False,
    ) -> None:
        """
        Grant table privileges to users in PostgreSQL

        :param schema: schema contains table to grant
        :param table: table name to grant
        :param list_users: list of users to grant
            if value is 'da', all members of DA team will be granted
        :param privileges: list of privileges to grant
            defaults to ['SELECT']
        :param all_privileges: whether to grant all privileges
            to users
            defaults to False
        """

        if not self.external_monitor:
            external_monitor = MonitorUserExternal(
                conn_username=self.db_user,
                conn_password=self.db_password,
                conn_host=self.host,
                conn_db=self.db,
                conn_number=1,
            )
            external_monitor.grant_table(
                schema=schema,
                table=table,
                list_users=list_users,
                privileges=privileges,
                all_privileges=all_privileges,
            )
        else:
            self.external_monitor.grant_table(
                schema=schema,
                table=table,
                list_users=list_users,
                privileges=privileges,
                all_privileges=all_privileges,
            )

    def _serialize_cell(
        self,
        cell,
    ) -> str:
        if cell is None:
            return None
        if isinstance(cell, datetime.datetime):
            return cell.isoformat()

        return str(cell)

    def read_sql_file(
        self,
        sql_file_path: str,
    ) -> str:
        """
        Get the SQL query given by SQL file

        :param sql_file_path: the located path of SQL file

        :return: SQL query
        """

        with open(sql_file_path, 'r') as file:
            sql = file.read()

        return sql

    def insert_data(
        self,
        data: pd.DataFrame,
        schema: str,
        table: str,
        ignore_errors: bool = False,
        commit_every: int = 1000,
        db_pool_conn: Callable = None,
    ) -> None:
        """
        Insert data to PostgreSQL table

        :param data: a dataframe contains data to insert
        :param schema: schema contains table to insert
        :param table: table name to insert
        :param ignore_errors: whether to ignore errors when inserting data
        :param commit_every: number rows of data to commit each time
            defaults to 1000
        :param db_pool_conn: connection pool contains multiple connections
            to connect to database
            defaults to None
            if None, a new connection will be created and automatically closed
            after being used
        """

        if not db_pool_conn:
            conn = self.open_conn()
        else:
            conn = db_pool_conn.getconn()

        try:
            cur = conn.cursor()

            target_fields = data.columns.to_list()
            target_fields = ", ".join(target_fields)
            target_fields = "({})".format(target_fields)

            rows = [
                tuple(r) for r in data.to_numpy()
            ]

            for i, row in enumerate(rows, 1):
                lst = []
                for cell in row:
                    lst.append(self._serialize_cell(cell))
                values = tuple(lst)
                placeholders = ["%s", ] * len(values)
                sql = "INSERT INTO "
                sql += "{0} {1} VALUES ({2})".format(
                    f'{schema}.{table}',
                    target_fields,
                    ",".join(placeholders)
                )
                if ignore_errors:
                    sql += " ON CONFLICT DO NOTHING"

                cur.execute(sql, values)
                if i % commit_every == 0:
                    conn.commit()
                    logging.info(
                        f"Loaded {i} into {table} rows so far"
                    )

            cur.close()
            conn.commit()

            if not db_pool_conn:
                self.close_conn(conn)
            else:
                db_pool_conn.putconn(conn)

            logging.info(
                f"Done loading. Loaded a total of {i} rows"
            )
        except Exception as e:
            if not db_pool_conn:
                self.close_conn(conn)
            else:
                db_pool_conn.putconn(conn)
            raise e

    def bulk_insert_data(
        self,
        data: pd.DataFrame,
        schema: str,
        table: str,
        commit_every: int = 1000,
        db_pool_conn: Callable = None,
    ) -> None:
        """
        Insert data to PostgreSQL table using COPY method
            (using when the data to insert is large)

        :param data: a dataframe contains data to insert
        :param schema: schema contains table to insert
        :param table: table name to insert
        :param commit_every: number rows of data to commit each time
            defaults to 1000
        :param db_pool_conn: connection pool contains multiple connections
            to connect to database
            defaults to None
            if None, a new connection will be created and automatically closed
            after being used
        """

        if len(data) == 0:
            return

        if not db_pool_conn:
            conn = self.open_conn()
        else:
            conn = db_pool_conn.getconn()

        try:
            cur = conn.cursor()

            columns = list(data.columns)

            data_chunk = [
                data[i:i+commit_every]
                for i in range(
                    0,
                    len(data),
                    commit_every,
                )
            ]

            for chunk in data_chunk:
                output = StringIO()
                chunk.to_csv(
                    output,
                    sep='|',
                    header=False,
                    index=False,
                )
                output.seek(0)

                copy_sql = f"COPY {schema}.{table} ({', '.join(columns)}) "
                copy_sql += "FROM stdin WITH CSV DELIMITER as '|'"

                cur.copy_expert(sql=copy_sql, file=output)
                conn.commit()
                logging.info(
                    f'Loaded {commit_every} into {table} rows so far'
                )

            cur.close()
            conn.commit()
            if not db_pool_conn:
                self.close_conn(conn)
            else:
                db_pool_conn.putconn(conn)

            logging.info(
                f"Done loading. Loaded a total of {len(data)} rows"
            )
        except Exception as e:
            if not db_pool_conn:
                self.close_conn(conn)
            else:
                db_pool_conn.putconn(conn)
            raise e

    def upsert_data(
        self,
        data: pd.DataFrame,
        schema: str,
        table: str,
        commit_every: int = 1000,
        db_pool_conn: Callable = None,
    ) -> None:
        """
        Upsert data (update the row if it already exist when inserting)
            to PostgreSQL table

        :param data: a dataframe contains data to upsert
        :param schema: schema contains table to upsert
        :param table: table name to upsert
        :param commit_every: number rows of data to commit each time
            defaults to 1000
        :param db_pool_conn: connection pool contains multiple connections
            to connect to database
            defaults to None
            if None, a new connection will be created and automatically closed
            after being used
        """

        if not db_pool_conn:
            conn = self.open_conn()
        else:
            conn = db_pool_conn.getconn()

        try:
            primary_keys = self.get_data(
                sql=f"""
                SELECT
                    a.attname,
                    format_type(a.atttypid, a.atttypmod) AS data_type
                FROM
                    pg_index i
                    JOIN pg_attribute a
                        ON a.attrelid = i.indrelid
                        AND a.attnum = ANY(i.indkey)
                WHERE
                    i.indrelid = '{schema}.{table}'::regclass
                    AND i.indisprimary
                """,
                db_pool_conn=db_pool_conn,
            )['attname'].to_list()

            cur = conn.cursor()

            target_fields = data.columns.to_list()
            target_fields = ", ".join(target_fields)
            target_fields = "({})".format(target_fields)

            to_update_columns = [
                col for col in data.columns
                if col not in primary_keys
            ]
            to_update_columns_sql = list()
            for col in to_update_columns:
                to_update_columns_sql.append(
                    f"{col} = excluded.{col}"
                )

            data = data.drop_duplicates().reset_index(drop=True)
            rows = [
                tuple(r) for r in data.to_numpy()
            ]

            for i, row in enumerate(rows, 1):
                lst = []
                for cell in row:
                    lst.append(self._serialize_cell(cell))
                values = tuple(lst)
                placeholders = ["%s", ] * len(values)
                sql = "INSERT INTO "
                sql += "{0} {1} VALUES ({2})".format(
                    f'{schema}.{table}',
                    target_fields,
                    ",".join(placeholders)
                )
                sql += " ON CONFLICT "
                sql += f"({', '.join(primary_keys)}) "
                sql += "DO UPDATE SET "
                sql += f"{', '.join(to_update_columns_sql)}"

                cur.execute(sql, values)
                if i % commit_every == 0:
                    conn.commit()
                    logging.info(
                        f"Loaded {i} into {table} rows so far"
                    )

            cur.close()
            conn.commit()
            if not db_pool_conn:
                self.close_conn(conn)
            else:
                db_pool_conn.putconn(conn)

            logging.info(
                f"Done loading. Loaded a total of {i} rows"
            )
        except Exception as e:
            if not db_pool_conn:
                self.close_conn(conn)
            else:
                db_pool_conn.putconn(conn)
            raise e

    def bulk_upsert_data(
        self,
        data: pd.DataFrame,
        schema: str,
        table: str,
        commit_every: int = 1000,
        db_pool_conn: Callable = None,
    ) -> None:
        """
        Upsert data (update the row if it already exist when inserting)
            to PostgreSQL table using COPY method
            (using when the data to uspert is large)

        :param data: a dataframe contains data to upsert
        :param schema: schema contains table to upsert
        :param table: table name to upsert
        :param commit_every: number rows of data to commit each time
            defaults to 1000
        :param db_pool_conn: connection pool contains multiple connections
            to connect to database
            defaults to None
            if None, a new connection will be created and automatically closed
            after being used
        """

        if len(data) == 0:
            return

        if not db_pool_conn:
            conn = self.open_conn()
        else:
            conn = db_pool_conn.getconn()

        try:
            primary_keys = self.get_data(
                sql=f"""
                SELECT
                    a.attname,
                    format_type(a.atttypid, a.atttypmod) AS data_type
                FROM
                    pg_index i
                    JOIN pg_attribute a
                        ON a.attrelid = i.indrelid
                        AND a.attnum = ANY(i.indkey)
                WHERE
                    i.indrelid = '{schema}.{table}'::regclass
                    AND i.indisprimary
                """,
                db_pool_conn=db_pool_conn,
            )['attname'].to_list()

            cur = conn.cursor()

            columns = list(data.columns)
            to_update_columns = [
                col for col in columns
                if col not in primary_keys
            ]
            to_update_columns_sql = list()
            for col in to_update_columns:
                to_update_columns_sql.append(
                    f"{col} = excluded.{col}"
                )

            tmp_table = f'tmp_{table}_bulk_upsert'
            create_tmp_table_sql = f"CREATE TEMP TABLE {tmp_table} "
            create_tmp_table_sql += "ON COMMIT DROP "
            create_tmp_table_sql += f"AS SELECT * FROM {schema}.{table} "
            create_tmp_table_sql += "WITH NO DATA"

            data = data.drop_duplicates().reset_index(drop=True)
            data_chunk = [
                data[i:i+commit_every]
                for i in range(
                    0,
                    len(data),
                    commit_every,
                )
            ]

            for chunk in data_chunk:
                cur.execute(create_tmp_table_sql)

                output = StringIO()
                chunk.to_csv(
                    output,
                    sep='|',
                    header=False,
                    index=False,
                )
                output.seek(0)

                copy_sql = f"COPY {tmp_table} ({', '.join(columns)}) "
                copy_sql += "FROM stdin WITH CSV DELIMITER as '|'"

                cur.copy_expert(sql=copy_sql, file=output)

                upsert_data_sql = f"INSERT INTO {schema}.{table} "
                upsert_data_sql += f"(SELECT * FROM {tmp_table}) "
                upsert_data_sql += "ON CONFLICT "
                upsert_data_sql += f"({', '.join(primary_keys)}) "
                upsert_data_sql += "DO UPDATE SET "
                upsert_data_sql += f"{', '.join(to_update_columns_sql)}"

                cur.execute(upsert_data_sql)

                conn.commit()
                logging.info(
                    f'Loaded {commit_every} into {table} rows so far'
                )

            cur.close()
            conn.commit()
            if not db_pool_conn:
                self.close_conn(conn)
            else:
                db_pool_conn.putconn(conn)

            logging.info(
                f"Done loading. Loaded a total of {len(data)} rows"
            )
        except Exception as e:
            if not db_pool_conn:
                self.close_conn(conn)
            else:
                db_pool_conn.putconn(conn)
            raise e

    def update_table(
        self,
        data: pd.DataFrame,
        schema: str,
        table: str,
        columns: list,
        commit_every: int = 1000,
        db_pool_conn: Callable = None,
    ) -> None:
        """
        Update new data of specific columns in the table
            based on primary keys

        :param data: a dataframe contains data to update
            (including primary keys and columns to update)
        :param schema: schema contains table to update data
        :param table: table to update data
        :param columns: list of column names to update data
        :param commit_every: number rows of data to commit each time
            defaults to 1000
        :param db_pool_conn: connection pool contains multiple connections
            to connect to database
            defaults to None
            if None, a new connection will be created and automatically closed
            after being used
        """

        if len(data) == 0:
            return

        if not db_pool_conn:
            conn = self.open_conn()
        else:
            conn = db_pool_conn.getconn()

        try:
            primary_keys = self.get_data(
                sql=f"""
                SELECT
                    a.attname,
                    format_type(a.atttypid, a.atttypmod) AS data_type
                FROM
                    pg_index i
                    JOIN pg_attribute a
                        ON a.attrelid = i.indrelid
                        AND a.attnum = ANY(i.indkey)
                WHERE
                    i.indrelid = '{schema}.{table}'::regclass
                    AND i.indisprimary
                """,
                db_pool_conn=db_pool_conn,
            )['attname'].to_list()

            cur = conn.cursor()

            sql = f"UPDATE {schema}.{table} SET "
            sql += ", ".join(
                [
                    f"{col} = %s" for col in columns
                ]
            )
            sql += " WHERE "
            sql += " AND ".join(
                [
                    f"{key} = %s" for key in primary_keys
                ]
            )

            sub_cols = columns + primary_keys
            rows = [
                tuple(r) for r in data[sub_cols].to_numpy()
            ]

            for i, row in enumerate(rows, 1):
                lst = []
                for cell in row:
                    lst.append(cell)
                values = tuple(lst)
                cur.execute(sql, values)
                if i % commit_every == 0:
                    conn.commit()
                    logging.info(
                        f"Updated {i} to {table} rows so far"
                    )

            cur.close()
            conn.commit()
            if not db_pool_conn:
                self.close_conn(conn)
            else:
                db_pool_conn.putconn(conn)

            logging.info(
                f"Done updating. Updated a total of {i} rows"
            )
        except Exception as e:
            if not db_pool_conn:
                self.close_conn(conn)
            else:
                db_pool_conn.putconn(conn)
            raise e

    def get_data(
        self,
        sql: str,
        db_pool_conn: Callable = None,
    ) -> pd.DataFrame:
        """
        Get data from PostgreSQL database given by a SQL query

        :param sql: the SQL query to get data
        :param db_pool_conn: connection pool contains multiple connections
            to connect to database
            defaults to None
            if None, a new connection will be created and automatically closed
            after being used

        :return: dataframe contains data extracted by the given SQL query
        """

        if not db_pool_conn:
            conn = self.open_conn()
        else:
            conn = db_pool_conn.getconn()

        try:
            data = pd.read_sql(
                sql=sql,
                con=conn,
            )

            if not db_pool_conn:
                self.close_conn(conn)
            else:
                db_pool_conn.putconn(conn)

            return data
        except Exception as e:
            if not db_pool_conn:
                self.close_conn(conn)
            else:
                db_pool_conn.putconn(conn)
            raise e

    def select_distinct(
        self,
        col: str,
        schema: str,
        table: str,
        db_pool_conn: Callable = None,
    ) -> list:
        """
        Get the distinct values of a specified column in a PostgreSQL table

        :param col: the column name to get the distinct data
        :param schema: the schema contains table to get data
        :param table: the table to get data
        :param db_pool_conn: connection pool contains multiple connections
            to connect to database
            defaults to None
            if None, a new connection will be created and automatically closed
            after being used

        :return: list of distinct data
        """

        if not db_pool_conn:
            conn = self.open_conn()
        else:
            conn = db_pool_conn.getconn()

        try:
            sql = f"SELECT DISTINCT {col} "
            sql += f"FROM {schema}.{table}"

            data = pd.read_sql(
                sql=sql,
                con=conn,
            )[col].to_list()

            if not db_pool_conn:
                self.close_conn(conn)
            else:
                db_pool_conn.putconn(conn)

            return data
        except Exception as e:
            if not db_pool_conn:
                self.close_conn(conn)
            else:
                db_pool_conn.putconn(conn)
            raise e

    def show_columns(
        self,
        schema: str,
        table: str,
        db_pool_conn: Callable = None,
    ) -> list:
        """
        Get list of columns name of a specific PostgreSQL table

        :param schema: the schema contains table to get columns
        :param table: the table to get columns
        :param db_pool_conn: connection pool contains multiple connections
            to connect to database
            defaults to None
            if None, a new connection will be created and automatically closed
            after being used

        :return: list of column names of the table
        """

        sql = "SELECT column_name "
        sql += "FROM information_schema.columns "
        sql += f"WHERE table_schema = {schema} "
        sql += f"AND table_name = {table}"

        columns = self.get_data(
            sql=sql,
            db_pool_conn=db_pool_conn,
        )['column_name'].to_list()

        return columns

    def execute(
        self,
        sql: str,
        fetch_output: bool = False,
        db_pool_conn: Callable = None,
    ) -> list | None:
        """
        Execute the given SQL query

        :param sql: SQL query to execute
        :param fetch_output: whether to fetch the results of the query
        :param db_pool_conn: connection pool contains multiple connections
            to connect to database
            defaults to None
            if None, a new connection will be created and automatically closed
            after being used

        :return: list of query output if fetch_output is True,
            otherwise None
        """

        if not db_pool_conn:
            conn = self.open_conn()
        else:
            conn = db_pool_conn.getconn()

        output = None

        try:
            cur = conn.cursor()

            cur.execute(sql)

            if fetch_output:
                output = cur.fetchall()

            cur.close()
            conn.commit()
            if not db_pool_conn:
                self.close_conn(conn)
            else:
                db_pool_conn.putconn(conn)
        except Exception as e:
            if not db_pool_conn:
                self.close_conn(conn)
            else:
                db_pool_conn.putconn(conn)
            raise e

        return output

    def add_column(
        self,
        schema: str,
        table: str,
        column_name: str | None = None,
        dtype: str | None = None,
        muliple_columns: dict = {},
        db_pool_conn: Callable = None,
    ) -> None:
        """
        Create new column for a specific table

        :param schema: the schema contains table to create column
        :param table: the table to create column
        :param column_name: name of the column to create
            available when creating single column
            defaults to None
        :param dtype: data type of the column to create
            available when creating single column
            defaults to None
        :param muliple_columns: dictionary contains columns name as key
            and data type of columns as value respectively
            defaults to empty dictionary
        :param db_pool_conn: connection pool contains multiple connections
            to connect to database
            defaults to None
            if None, a new connection will be created and automatically closed
            after being used
        """

        sql = f"ALTER TABLE {schema}.{table} "
        if muliple_columns:
            sql += ", ".join(
                [
                    f"ADD COLUMN IF NOT EXISTS {col} {muliple_columns[col]}"
                    for col in muliple_columns.keys()
                ]
            )
        else:
            sql += f"ADD COLUMN IF NOT EXISTS {column_name} {dtype}"

        self.execute(
            sql=sql,
            db_pool_conn=db_pool_conn,
        )

    def create_table(
        self,
        schema: str,
        table: str,
        columns_with_dtype: dict,
        columns_primary_key: list = [],
        columns_not_null: list = [],
        columns_with_default: dict = {},
        db_pool_conn: Callable = None,
    ) -> None:
        """
        Create new table in the database

        :param schema: schema contains table to create
        :param table: table name to create
        :param columns_with_dtype: dictionary contains column names
            as key and the data type of column as value respectively
        :param columns_primary_key: list of columns to set primary keys
            defaults to empty list
        :param columns_not_null: list of columns to set constraints not null
            defaults to empty list
        :param columns_with_default: dictionary contains column names
            as key and the default value of column as value respectively
        :param db_pool_conn: connection pool contains multiple connections
            to connect to database
            defaults to None
            if None, a new connection will be created and automatically closed
            after being used
        """

        cols_sql = list()
        for col in columns_with_dtype.keys():
            col_sql = f"{col} {columns_with_dtype[col]}"
            if col in columns_not_null:
                col_sql += " NOT NULL"
            if col in columns_with_default.keys():
                if isinstance(columns_with_default[col], str):
                    default_value = f"'{columns_with_default[col]}'"
                else:
                    default_value = columns_with_default[col]
                col_sql += f" DEFAULT {default_value}"
            cols_sql.append(col_sql)
        if columns_primary_key:
            cols_sql.append(
                f"PRIMARY KEY ({', '.join(columns_primary_key)})"
            )

        sql = f"CREATE TABLE IF NOT EXISTS {schema}.{table} "
        sql += f"({', '.join(cols_sql)})"

        self.execute(
            sql=sql,
            db_pool_conn=db_pool_conn,
        )

    def generate_log_data_path(
        self,
        file_path: str,
        database: str,
        mode: str,
        schema: str = None,
        table: str = None,
        owner: str = None,
        transform_func: str = None,
        db_pool_conn: Callable = None,
    ) -> None:
        """
        Generates log of data path when ingest raw data to minIO

        :param file_path: path to raw data file in minIO
        :param database: database platform where the processed data
            will be stored
        :param mode: the storage mode where the raw data is located in minIO
            the value accepted is 'prod' or 'stag'
        :param schema: schema where the processed data will be stored
            defaults to None
        :param table: table where the processed data will be stored
            defaults to None
        :param owner: name of owner belonging to the raw data path
            defaults to None
        :param transform_func: name of the transform function that
            is used to transform the raw data
            defaults to None
        :param db_pool_conn: connection pool contains multiple connections
            to connect to database
            defaults to None
            if None, a new connection will be created and automatically closed
            after being used
        """

        if mode == 'prod':
            df = pd.DataFrame(
                {
                    'log_date': str(
                        datetime.datetime.now().replace(microsecond=0)
                        + datetime.timedelta(hours=7)
                    ),
                    'db_type': database,
                    'schema': schema,
                    'table_name': table,
                    'file_path': file_path,
                    'status': 0,
                    'owner': owner,
                    'transform_func': transform_func,
                },
                index=[0],
            )

            self.upsert_data(
                data=df,
                schema='metadata',
                table='platform_log_raw_data',
                primary_keys=['log_date', 'file_path'],
                db_pool_conn=db_pool_conn,
            )

            logging.info('Done generating log data path')

    def update_log_data_path(
        self,
        file_path: str,
        log_date: str,
        db_pool_conn: Callable = None,
    ) -> None:
        """
        Update the status of the log of raw data path to success
            to tracking whether the raw data is processed
            and load to the database

        :param file_path: path to raw data file in minIO
        :param log_date: the date that the log was generated
            the date format will be 'yyyy-mm-dd'
        :param db_pool_conn: connection pool contains multiple connections
            to connect to database
            defaults to None
            if None, a new connection will be created and automatically closed
            after being used
        """

        df = pd.DataFrame(
            {
                'log_date': log_date,
                'file_path': file_path,
                'status': 1,
            },
            index=[0],
        )

        self.update_table(
            data=df,
            schema='metadata',
            table='platform_log_raw_data',
            columns=['status'],
            db_pool_conn=db_pool_conn,
        )

        logging.info('Done updating log data path')

    def get_unload_data_path(
        self,
        transform_func: str,
        db_pool_conn: Callable = None,
    ) -> pd.DataFrame:
        """
        Get all the raw data path the is not processed
            and load to the database in the current date
            based on the name of the transform function

        :param transform_func: name of the transform function
            to process the raw data
        :param db_pool_conn: connection pool contains multiple connections
            to connect to database
            defaults to None
            if None, a new connection will be created and automatically closed
            after being used

        :return: a dataframe contains the information
            about the log raw data path
        """

        current_date = str(
            (
                datetime.datetime.now().replace(microsecond=0)
                + datetime.timedelta(hours=7)
            ).date()
        )
        sql = "SELECT * FROM metadata.platform_log_raw_data "
        sql += f"WHERE (CAST(log_date as DATE) = '{current_date}' "
        sql += f"AND transform_func = '{transform_func}' "
        sql += "AND status = 0)"

        df = self.get_data(
            sql=sql,
            db_pool_conn=db_pool_conn,
        )
        df.sort_values(
            by=['log_date'],
            ascending=True,
            inplace=True,
        )

        logging.info(f'Total {len(df)} unload data path in {current_date}')

        return df

    def truncate_table(
        self,
        schema: str,
        table: str,
        reset_identity: bool = False,
        db_pool_conn: Callable = None,
    ) -> None:
        """
        Remove all the data of the table

        :param schema: schema contains table to truncate
        :param table: table name to truncate
        :param reset_identity: whether to reset identity of the table
            defaults to False
        :param db_pool_conn: connection pool contains multiple connections
            to connect to database
            defaults to None
            if None, a new connection will be created and automatically closed
            after being used
        """

        sql = f"TRUNCATE TABLE {schema}.{table}"
        if reset_identity:
            sql += " RESTART IDENTITY"

        self.execute(
            sql=sql,
            db_pool_conn=db_pool_conn,
        )

    def table_exists(
        self,
        schema: str,
        table: str,
        db_pool_conn: Callable = None,
    ) -> bool:
        """
        Check if the table exists in database

        :param schema: schema contains table to check
        :param table: table name to check
        :param db_pool_conn: connection pool contains multiple connections
            to connect to database
            defaults to None
            if None, a new connection will be created and automatically closed
            after being used

        :return: True if table exists and False if not
        """

        sql = "SELECT EXISTS (SELECT FROM pg_tables WHERE "
        sql += f"schemaname = '{schema}' AND tablename  = '{table}')"

        exists = self.get_data(
            sql=sql,
            db_pool_conn=db_pool_conn,
        )['exists'].values[0]

        return exists
